/**********************************************/
/*                                            */
/*  例題8.2 2次元ポアソン方程式の差分法のプログラム  */
/*                                            */
/**********************************************/

#include<stdio.h>
#include<math.h>

#define I (10)//xに関する分割数
#define J (10)//yに関する分割数

#define f(X,Y) ( (cos(M_PI*(X)/2.0))*((1.0+(Y))*(1.0+(Y))*(1.0+(Y))*(1.0-(Y))*M_PI*M_PI/4.0+((1.0+(Y))*12.0*(Y))) )

int main(void)
{
    /*計算で必要な変数*/
    int i, j, k;
    double A[(I-1)*(J-1)][(I-1)*(J-1)],u[(I-1)*(J-1)],B[(I-1)*(J-1)],F[(I-1)*(J-1)];
    double L[(I-1)*(J-1)][(I-1)*(J-1)],R[(I-1)*(J-1)][(I-1)*(J-1)];
    double dx,dy,x,y,sum,max1,max2,old_max,value,U,omega;
	
    /*空間をx軸に関してI等分, y軸に関してJ等分する*/
	dx=2.0/I;
    dy=2.0/J;
    
    /*左辺の行列Aの設定*/
    for(i=0;i<(I-1)*(J-1);i++)
    {
        for(j=0;j<(I-1)*(J-1);j++)
        {
            A[i][j]=0.0;
        }
    }
    
    for(j=0;j<(I-1)*(J-1);j++)
    {
        if(I-1+j<(I-1)*(J-1))
        {
            A[I-1+j][j]=-1.0/(dy*dy);
            A[j][I-1+j]=-1.0/(dy*dy);
        }
        
        A[j][j]=2.0/(dx*dx)+2.0/(dy*dy);
        
        A[j][j+1]=-1.0/(dx*dx);
        A[j+1][j]=-1.0/(dx*dx);
        
        if(j%(I-1)==I-2)
        {
            A[j][j+1]=0.0;
        }
        if((j+1)%(I-1)==0)
        {
            A[j+1][j]=0.0;
        }
    }
    
    /*右辺のベクトルf+bの設定*/
    for(i=0;i<I-1;i++)
    {
        for(j=0;j<J-1;j++)
        {
            B[j*(I-1)+i]=f(-1.0+(i+1)*dx,-1.0+(j+1)*dy);
        }
    }
    
    /*SOR法のための設定*/
    /*初期値の設定*/
    for(i=0;i<(I-1)*(J-1);i++)
    {
        u[i]=0.0;
    }

    /*omegaの値の設定*/
    omega=1.94;
    
    /*SOR法の左辺 D/omega+L の設定*/
    for(i=0;i<(I-1)*(J-1);i++)
    {
        for(j=0;j<(I-1)*(J-1);j++)
        {
            L[i][j]=0.0;
            if(j<i)
            {
                L[i][j]=A[i][j];
            }
            if(i==j)
            {
                L[i][j]=A[i][j]/omega;
            }
        }
    }
    
    /*SOR法の右辺 (1-1/omega)D+U の設定*/
    for(i=0;i<(I-1)*(J-1);i++)
    {
        for(j=0;j<(I-1)*(J-1);j++)
        {
            R[i][j]=0.0;
            if(i<j)
            {
                R[i][j]=A[i][j];
            }
            if(i==j)
            {
                R[i][j]=A[i][j]*(1.0-1.0/omega);
            }
        }
    }
    
    /*SOR法による反復*/
    for(k=0;k<=20000;k++)
    {
        for(i=0;i<(I-1)*(J-1);i++)
        {
            F[i]=0.0;
            for(j=0;j<(I-1)*(J-1);j++)
            {
                F[i]=F[i]-R[i][j]*u[j];
            }
            F[i]=F[i]+B[i];
        }

        u[0]=F[0]/L[0][0];
        
        for(i=1;i<(I-1)*(J-1);i++)
        {
            sum=0.0;
            for(j=0;j<=i-1;j++)
            {
                sum=sum+L[i][j]*u[j];
            }
            u[i]=(F[i]-sum)/L[i][i];
        }
        
        /*相対誤差の計算*/
        max1=0.0;
        max2=0.0;
        for(i=0;i<I-1;i++)
        {
            for(j=0;j<J-1;j++)
            {
                U=cos(M_PI*(-1.0+dx*(i+1))/2.0)*((dy*(j+1))*(dy*(j+1))*(dy*(j+1))*(1.0-(-1.0+dy*(j+1))));
                value=(u[j*(I-1)+i]-U);
                
                if(max1<fabs(value))
                {
                    max1=fabs(value);
                }
                
                if(max2<fabs(U))
                {
                    max2=fabs(U);
                }
            }
        }
        
        /*結果の表示*/
        printf("%d %e %f\n",k,max1/max2,(max1/max2)/(dx*dx));
        
        /*計算が収束すると終了*/
        if(fabs(old_max-max1)<0.00000000001)
        {
            break;
        }
        old_max=max1;
    }
}


