/*****************************************************/
/*                                                   */
/*  例題8.3 熱方程式をクランク・ニコルソン法で解くプログラム  */
/*                                                   */
/*****************************************************/

#include<stdio.h>
#include<math.h>

#define N (100)//分割数の設定
#define T (0.5)
#define f(X,T) (exp(-M_PI*M_PI*(T)/4.0)*sin(((X)+1.0)*M_PI/2.0))

int main(void)
{
    /*計算で必要な変数*/
    int i, j, k, l, TMAX;
	double ori_A[N-1][N-1],A[N-1][N],u[N-1],new_u[N-1],r,dx,dt,t,max1,max2,U,sum;
	
    /*区間をN分割する*/
	dx=2.0/N;
    r=0.4;//rの値の設定
    dt=r*dx*dx;
    TMAX=T/dt;
    
    /*初期値の設定*/
    for(i=0;i<N-1;i++)
    {
        u[i]=sin((-1.0+(i+1)*dx+1.0)*M_PI/2.0);
    }
    
    /*時刻後退差分法の左辺の行列の設定*/
    for(i=0;i<N-1;i++)
    {
        for(j=0;j<N-1;j++)
        {
            ori_A[i][j]=0.0;
        }
    }
    
    for(i=0;i<N-1;i++)
    {
        ori_A[i][i]=1.0/dt+1.0/(dx*dx);
        if(i+1<N-1)
        {
            ori_A[i][i+1]=-0.5/(dx*dx);
            ori_A[i+1][i]=-0.5/(dx*dx);
        }
    }
    
    /*時刻後退差分による時間発展*/
    for(l=0;l<=TMAX;l++)
    {
        t=l*dt;
        
        /*行列Aに行列ori_Aを代入*/
        for(i=0;i<N-1;i++)
        {
            for(j=0;j<N-1;j++)
            {
                A[i][j]=ori_A[i][j];
            }
        }
        
        /*行列AのN-1列目に右辺のベクトルを代入*/
        for(i=1;i<N-2;i++)
        {
            A[i][N-1]=u[i]/dt+0.5*(u[i+1]-2.0*u[i]+u[i-1])/(dx*dx);
        }
        A[0][N-1]=u[0]/dt+0.5*(u[1]-2.0*u[0])/(dx*dx);
        A[N-2][N-1]=u[N-2]/dt+0.5*(-2.0*u[N-2]+u[N-3])/(dx*dx);
        
        /*ガウスの消去法 前進過程*/
        for(i=0;i<N-2;i++)
        {
            for(j=1+i;j<N-1;j++)
            {
                for(k=1+i;k<N;k++)
                {
                    A[j][k]=A[j][k]-A[j][i]*A[i][k]/A[i][i];
                }
            }
        }
        
        /*ガウスの消去法 後進過程*/
        u[N-2]=A[N-2][N-1]/A[N-2][N-2];
        for(i=N-3;i>=0;i--)
        {
            sum=0.0;
            for(j=i+1;j<=N-2;j++)
            {
                sum=sum+u[j]*A[i][j];
            }
            u[i]=(A[i][N-1]-sum)/A[i][i];
        }
    }
    
    /*結果の表示*/
    max1=0.0;
    max2=0.0;
    for(i=0;i<N-1;i++)
    {
        U=f(-1.0+(i+1)*dx,t+dt);
        if(max1<fabs(u[i]-U))
            max1=fabs(u[i]-U);
        if(max2<fabs(U))
            max2=fabs(U);
    }
    printf("%e\n",max1/max2);
}


